package edu.rice.comp211.laundry.list;

import edu.rice.comp211.laundry.iterator.BiIteratorI;
import edu.rice.comp211.laundry.iterator.IteratorException;
import edu.rice.comp211.laundry.iterator.IteratorI;

/**
 * A circularly linked class implementing the BiListI interface.
 */
public class BiList<T> implements BiListI<T> {

	/**
	 * The header node for the list.
	 */
	private Node<T> head = new Node<T>();
	/**
	 * The length of this list.
	 */
	private int length = 0;

	/* Relying on the default constructor for BiList. */

	@Override
	public String toString() {
		StringBuilder s = new StringBuilder("(");
		for (IteratorI<T> i = newIterator(); !i.atEnd(); i.next()) {
			if (i.atStart()) {
				s.append(i.currentItem().toString());
			} else {
				s.append(" " + i.currentItem().toString());
			}
		}
		return s.append(")").toString();
	}

	/* Implementations of BiListI methods.  See BiListI for contracts. */
	public boolean isEmpty() {
		return head.succ == head;
	}

	public <R> R accept(BiListIVisitor<T, R> visitor){
		if(head.succ == head) return visitor.forEmpty(this);
		else return visitor.forNonEmpty(this);
	}

	/**
	 * Factory method for a new, empty list that holds the same type of elements as this list.
	 * @return a new BiList&lt;T&gt;
	 */
	public BiListI<T> newList() {
		return new BiList<T>();
	}

	/**
	 * Returns the number of elements in the list
	 * @return the number of elements in the list
	 */
	public int length() {
		return length;
	}

	/**
	 * Inserts the given value at the front of the list
	 * @param o the new element to insert
	 * @return a reference to this list.
	 */
	public BiList<T> insertFront(T o) {
		Node<T> next = head.succ;
		Node<T> newNode = new Node<T>(o, head, next);  // allocate new Node
		// insert new Node
		head.succ = newNode;
		next.pred = newNode;
		length++;
		return this;
	}

	/**
	 * Insert the given value at the end of the list
	 * @param o the new element to insert
	 * @return a reference to this list
	 */
	public BiList<T> insertRear(T o) {
		Node<T> prev = head.pred;
		Node<T> newNode = new Node<T>(o, prev, head);  // allocate new Node
		// insert new Node
		head.pred = newNode;
		prev.succ = newNode;
		length++;
		return this;
	}

	/**
	 * Removes the first element of the list
	 * @return the element that was removed.
	 * @throws IllegalState exception if list is empty
	 */
	public T remFront() {
		if (isEmpty()) {
			throw new IllegalStateException("Attempt to remove first element of empty BiList");
		}
		Node<T> next = head.succ;
		T remOb = next.item;
		Node<T> newNext = next.succ;
		head.succ = newNext;
		newNext.pred = head;
		length--;

		return remOb;
	}

	/**
	 *  Factory method that creates an iterator initialized to point at the start of the list.
	 *  Virtual constructors for iterators; identical except for return type 
	 *  @return a BiIteratorI&lt;T&gt; interator
	 */
	public BiIteratorI<T> newIterator() {
		return new BiIterator();
	}

	/**
	 * Remove the last element in the list
	 * @return the element that was removed
	 * @throws IllegalStateException if the list is empty.
	 */
	public T remRear() {
		if (isEmpty()) {
			throw new IllegalStateException("Attempt to remove last element of empty BiList");
		}
		Node<T> last = head.pred;
		T remOb = last.item;
		Node<T> lastPred = last.pred;
		head.pred = lastPred;
		lastPred.succ = head;
		length--;
		return remOb;
	}

	/**
	 * Returns true if all elements in this list and the given list are equal.
	 * @param other the other list
	 * @return true if all elements are equal, false otherwise.
	 */
	public boolean equals(Object other) {
		if (other == null || getClass() != other.getClass()) {
			return false;
		}
		@SuppressWarnings("unchecked")
		BiList<T> b = (BiList<T>) other;
		if (length != b.length) {
			return false;
		}
		IteratorI<T> i = newIterator();
		IteratorI<T> j = b.newIterator();
		for (; !i.atEnd(); i.next(), j.next()) {
			if (!i.currentItem().equals(j.currentItem())) {
				return false;
			}
		}
		return true;
	}

	/**
	 *  Static nested Node class.  Embedded as nested class to hide access.
	 *  @param <T> the type of element being held 
	 */
	private static class Node<T> {
		/**
		 * The element held by this node
		 */
		T item;
		
		/**
		 * The predecessor and successor nodes
		 */
		Node<T> pred, succ;

		/**
		 * Constructor
		 * @param i the element held
		 * @param p the predecessor node
		 * @param s the successor node.
		 */
		Node(T i, Node<T> p, Node<T> s) {
			item = i;
			pred = p;
			succ = s;
		}

		/**
		 * Constructor that creates an empty, self-referencing node that is used to represent an empty list.
		 */
		Node() {  // allocate header
			item = null;
			pred = this;
			succ = this;
		}
	}

	/**
	 * A class used to generate iterators for this list.
	 * A genuine named inner class that includes a link (BiList.this) to the enclosing BiList.
	 */
	private class BiIterator implements BiIteratorI<T> {

		/**
		 * The current node the iterator is pointing to
		 */
		Node<T> current;

		/**
		 * Constructor for the class
		 * Initializes the current node to reference the front of the list
		 */
		BiIterator() {
			current = BiList.this.head.succ;
		} // current is first item (if one exists) }

		@Override
		public void first() {
			current = BiList.this.head.succ;
		}

		@Override
		public void last() {
			current = BiList.this.head.pred;  // current is last item (if one exists)
		}

		@Override
		public void next() {
			current = current.succ;        // wraps around end
		}
		@Override
		public void prev() {
			current = current.pred;        // wraps around end
		}
		@Override
		public T currentItem() {
			if (current == BiList.this.head) {
				throw new IteratorException("No current element in " + BiList.this);
			}
			return current.item;
		}
		@Override
		public boolean atEnd() {
			return current == BiList.this.head;
		}

		/**
		 * Returns true if BiList is non-empty and current is the first node in the list.
		 */
		public boolean atStart() {
			return !atEnd() && current == BiList.this.head.succ;
		}

		@Override
		public void insert(T o) {

			// pre: true
			// post: Node containing o is inserted before current item,
			//       current is unchanged

			Node<T> oldPred = current.pred;
			Node<T> newNode = new Node<T>(o, oldPred, current);  // allocate new node
			current.pred = newNode;                              // insert it
			oldPred.succ = newNode;
			BiList.this.length++;
		}

		@Override
		public T remove() {

			// pre:  current is valid
			// post: current becomes current.succ

			if (current == BiList.this.head) {
				throw new IteratorException("BiIterator.remove() applied at end of BiList " + BiList.this);
			}
			Node<T> cPred = current.pred;
			Node<T> cSucc = current.succ;
			cPred.succ = cSucc;
			cSucc.pred = cPred;
			T remOb = current.item;
			current = cSucc;
			BiList.this.length--;
			return remOb;
		}
	}
}
